/*
 * dsm.c - Unihedron DSM interface routines
 *
 * $Id: dsm.c,v 1.13 2008/08/30 06:11:11 johns Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"
#include "serial.h"
#include "dsm.h"

typedef struct dsmhandle_t {
  RS232CommHandle commhandle;     /* rs232 port */
  rt_timerhandle commtimer;       /* communications timer */
  unsigned char buf[1024];
  int bufpos;                     /* first unused char in buffer */
  int status;                     /* health of DSM comm channel */
  int lastcount;                  /* count value of last packet received */
  int droppedpackets;             /* count of packet sequence numbers missed */
  int serialno;                   /* serial number of received packet */

  int skyaduA;                    /* DDC112 sky adu count */
  int ddcrangeA;                  /* DDC112 range setting */
  int skyaduB;                    /* DDC112 sky adu count */
  int ddcrangeB;                  /* DDC112 range setting */
  int tempadu;                    /* Temperature ADU */
  int period;                     /* Integration period */
  int aperture;                   /* Aperture wheel position */
  int aperturemoving;             /* true if the wheel is in motion */
  int aperturetoggle;             /* last aperture toggle status */
  int apertureunknown;            /* true if the wheel position is unknown */
} dsmhandle;


int dsm_flushbuffer(DSMHandle voidhandle) {
  dsmhandle * handle = (dsmhandle *) voidhandle;
  while (rs232_comm_read(handle->commhandle, (char *) handle->buf, sizeof(handle->buf) - 1) > 0);
  handle->bufpos = 0; /* clear out any pending data */
  return DSM_ERR_NONE;
}


DSMHandle dsm_open(const char * dsmname, int highspeed) {
  dsmhandle * handle;
  int baudrate = (highspeed) ? 115200 : 9600;

  if (dsmname == NULL)
    return NULL;

  handle = (dsmhandle *) malloc(sizeof(dsmhandle));
  if (handle == NULL)
    return NULL;

  /* clear all values in dsmhandle to 0 */
  memset(handle, 0, sizeof(dsmhandle));

  if (rs232_comm_open(dsmname, &handle->commhandle, baudrate)) {
    printf("failed to open comm port at %d baud\n", baudrate);
    free(handle);
    return NULL;
  }

  handle->commtimer = rt_timer_create();
  rt_timer_start(handle->commtimer);

  dsm_flushbuffer(handle); /* flush any existing incoming data */

  return handle; /* successfull open */
}


int dsm_close(DSMHandle voidhandle) {
  dsmhandle * handle = (dsmhandle *) voidhandle;

  rt_timer_destroy(handle->commtimer);
  free(handle);

  return DSM_ERR_NONE;
}


static int dsm_update(DSMHandle voidhandle, int maxpacks) {
  int i, num;
  dsmhandle * handle = (dsmhandle *) voidhandle;

  int remainder = 0;
  int lastpackstart = -1;
  int packs = 0;
  int readsize = sizeof(handle->buf) - handle->bufpos*sizeof(char) - 1;

  if (handle == NULL)
    return DSM_ERR_BADHANDLE;

  packs = 0; /* no packs received yet */

  /* read available chars into empty portion of the buffer */
  num = rs232_comm_read(handle->commhandle, (char *) &handle->buf[handle->bufpos], readsize);
  handle->bufpos += num;

  /* process packets if we have enough characters in the buffer */
#if 1
  if (handle->bufpos > 8) {
#else
  if (handle->bufpos > (sizeof(handle->buf) / 2)) {
#endif
    for (i=0; (i<(handle->bufpos-8)) && (packs < maxpacks); i++) {
      int sl, count, cntdist, intmode, side, adutmp, ddcrangetmp, testmode; 
      unsigned int cksum;
      for (cksum=0,sl=i; sl<(i+8); sl++) 
        cksum+=handle->buf[sl];

      /* compare checksums */
      if ((cksum & 0xff) != handle->buf[i+8])
        continue; /* skip to next offset */

      /* if we made it this far, we got a valid checksum at offset i */
      lastpackstart = i;
      side        = (handle->buf[i] & 0x80) >> 7;
      testmode    = (handle->buf[i] & 0x40) >> 6;
      intmode     = (handle->buf[i] & 0x20) >> 5;
      adutmp      = ((handle->buf[i  ] & 0x0f) << 16) |
                     (handle->buf[i+1] << 8) | handle->buf[i+2];
      ddcrangetmp = handle->buf[i+3];

      if (testmode) {
        handle->status |= DSM_TEST;
      } 

      if (side) {
        handle->status |= DSM_SIDEA;
        handle->skyaduA = adutmp;
        handle->ddcrangeA = ddcrangetmp;
      } else {
        handle->status |= DSM_SIDEB;
        handle->skyaduB = adutmp;
        handle->ddcrangeB = ddcrangetmp;
      }

      /* only store most recent aperture wheel status */
      handle->aperture        = handle->buf[i+4] & 0x1f;
      handle->apertureunknown = handle->buf[i+4] & 0x80;
      handle->aperturetoggle  = handle->buf[i+4] & 0x40;
      handle->aperturemoving  = handle->buf[i+4] & 0x20;

      adutmp = (handle->buf[i+5] << 8) | handle->buf[i+6];
      if (intmode) {
        handle->status |= DSM_TEMP;
        handle->tempadu = adutmp;
      } else {
        handle->status |= DSM_PERIOD;
        handle->period  = adutmp;
      }

      count = handle->buf[i+7];
      cntdist = count - handle->lastcount;


      if (cntdist > 1) {
        if ((count != 0) && (handle->lastcount != 255)) { 
          handle->droppedpackets++;
          printf("** Dropped packet.  last count: %d  current count: %d\n",
               handle->lastcount, count); 
        }
      }
      handle->lastcount = count;

      /* process next packet */
      packs++;
      i+=8; /* loop post-increment catches the other 1 */
    }

    /* If we receive any valid packets, we've got a good  */
    /* link to the DSM and it is operating comm-wise.     */
    /* This can be used to distinguish between situations */
    /* where the DSM has no position fix, is reporting    */
    /* more slowly than expected, or is just plain dead.  */
    if (packs > 0) {
      handle->status |= DSM_COMM;
      handle->status |= DSM_DATA;
    }

    /* reset buffer fullness and copy remaining unused bytes  */
    remainder = handle->bufpos - (lastpackstart+9);
#if 0
    printf("lps: %d lps+8: %d bufpos: %d rem: %d\n", lastpackstart, lastpackstart+8, handle->bufpos, remainder);
#endif
    if (lastpackstart >= 0) {
      memmove(&handle->buf[0], &handle->buf[lastpackstart+9], remainder);
      handle->bufpos = remainder;
    } else {
      memmove(&handle->buf[0], &handle->buf[handle->bufpos-8], 8);
      handle->bufpos = 8;
    }

    /* increment count of valid packets */
    handle->serialno += packs;
  } 

  return packs;
}

int dsm_comm_write(DSMHandle voidhandle, const char *buf, int sz) {
  dsmhandle * handle = (dsmhandle *) voidhandle;
  int i; 
  char cbuf[2];
  cbuf[0] = '\0';
  cbuf[1] = '\0';

  /* pace the output so as not to overrun the DSM CPU */ 
  for (i=0; i<sz; i++) {
    cbuf[0] = buf[i];

    /* retry dropped chars */
    while (rs232_comm_write(handle->commhandle, cbuf) == 0);

    rt_sleep(1);
  }

  return DSM_ERR_NONE;   
}


int dsm_reset(DSMHandle voidhandle) {
  dsmhandle * handle = (dsmhandle *) voidhandle;

  int rc;
  char cbuf[2];
  cbuf[0] = 0x19;
  cbuf[1] = '\0';
  rc=dsm_comm_write(voidhandle, cbuf, 1);

  dsm_flushbuffer(handle); /* flush any existing incoming data */

  /* wait here until we begin getting data again or we timeout */
  dsm_syncup_status(handle);            /* eat all buffered packets */
  while (dsm_update(handle, 100) == 0); /* wait for new data */
  dsm_syncup_status(handle);            /* eat all buffered packets */
  while (dsm_update(handle, 100) == 0); /* wait for new data */

  return DSM_ERR_NONE;
}


int dsm_set_test_mode(DSMHandle voidhandle, int on) {
  char cbuf[16];
  sprintf(cbuf, "T%c\r", (on) ? '1' : '0');

  return dsm_comm_write(voidhandle, cbuf, 3);
}


int dsm_set_integration_count(DSMHandle voidhandle, int icount) {
  char cbuf[16];
  int cmdack, oldperiod, rc;
  dsmhandle * handle = (dsmhandle *) voidhandle;

  if ((icount < 10) || (icount > 9765)) 
    return DSM_ERR_BADPARAM;
  
  sprintf(cbuf, "I%04d\r", icount);

  oldperiod = handle->period;
  dsm_clear_status(voidhandle);       /* clear old status before updates */

  cmdack=0;
  while (!cmdack) {
    int try;
    rc=dsm_comm_write(voidhandle, cbuf, 6);

    if (rc < 0) {
      printf("Comm error!\n");
      return rc;
    }
 
    for (try=0; try<200; try++) {
      while (dsm_update(handle, 100) == 0); /* wait for new data */
      if (handle->period != oldperiod) {
        printf("Integration time command acknowledged.\n");
        cmdack=1;
        break;
      }
    }

    if (!cmdack)
      printf("Retrying integration time command...\n");
  }

  return DSM_ERR_NONE;
}


int dsm_set_integration_time(DSMHandle voidhandle, float time) {
  int icount = (int) (time / 0.0001024);

  return dsm_set_integration_count(voidhandle, icount);
}


int dsm_set_aperture(DSMHandle voidhandle, int aperture) {
  dsmhandle * handle = (dsmhandle *) voidhandle;
  char cmdstr[32];
  int movepos, oldtoggle, rc, cmdack, try;

  if (handle == NULL)
    return DSM_ERR_BADHANDLE;

  dsm_syncup_status(handle);            /* eat all buffered packets */
  while (dsm_update(handle, 100) == 0); /* wait for new data */

  oldtoggle = handle->aperturetoggle;   /* previous toggle status */
  dsm_clear_status(voidhandle);         /* clear old status before updates */

  handle->aperture = -1; /* set to invalid aperture before move command */

  switch(aperture) {
    case DSM_APERTURE_WHITE:
    case DSM_APERTURE_IRFILTER:
    case DSM_APERTURE_BLACK:
    case DSM_APERTURE_OPEN:
      movepos=aperture;
      break;

    default:
      return DSM_ERR_BADPARAM;
  }

  /* Send aperture move command, retrying until it is acknowledged */
  sprintf(cmdstr, "M%c\r", 0x40 + movepos);
  cmdack=0;
  while (!cmdack) {
    rc=dsm_comm_write(voidhandle, cmdstr, 3);

    if (rc < 0) {
      printf("Comm error!\n");
      return rc;
    }
 
    for (try=0; try<800; try++) {
      dsm_syncup_status(handle);            /* eat all buffered packets */
      while (dsm_update(handle, 100) == 0); /* wait for new data */
      if (handle->aperturetoggle != oldtoggle) {
        printf("Aperture move command acknowledged.\n");
        cmdack=1;
        break;
      }
    }

    if (!cmdack)
      printf("Retrying aperture wheel move command...\n");
  }

  /* loop waiting for DSM aperture wheel to reach */
  /* the commanded position                       */
  try=0;
  while (try == 0 || handle->aperture != aperture || 
         handle->aperturemoving || handle->apertureunknown) {
    int packs = dsm_update(handle, 1); /* check for new data */
    if (packs > 0) {
      dsm_print_aperture_status(voidhandle);
      try++;
    }
  }
  printf("\n");

  return DSM_ERR_NONE;
}


int dsm_clear_status(DSMHandle voidhandle) {
  dsmhandle * handle = (dsmhandle *) voidhandle;
  handle->status    = 0;
  handle->skyaduA   = 0;
  handle->ddcrangeA = 0;
  handle->skyaduB   = 0;
  handle->ddcrangeB = 0;
  handle->tempadu   = 0;
  handle->period    = 0;
  handle->aperture  = -1;
  handle->apertureunknown = 1;
  handle->aperturetoggle = 0;

  return DSM_ERR_NONE;
}


int dsm_syncup_status(DSMHandle voidhandle) {
  dsmhandle * handle = (dsmhandle *) voidhandle;
  int packs;

  if (handle == NULL)
    return DSM_ERR_BADHANDLE;

  /* loop until we've read all packets in the input buffer */
  while ((packs = dsm_update(handle, 100)) != 0); /* check for new data */

  return DSM_ERR_NONE;
}
  

int dsm_get_status(DSMHandle voidhandle) {
  dsmhandle * handle = (dsmhandle *) voidhandle;
  int packs;

  if (handle == NULL)
    return DSM_ERR_BADHANDLE;

  packs = dsm_update(handle, 1); /* check for new data */

  return packs;
}


int dsm_get_soonest_statuspair(DSMHandle voidhandle) {
  dsmhandle * handle = (dsmhandle *) voidhandle;
  int packs, totalpacks;

  if (handle == NULL)
    return DSM_ERR_BADHANDLE;

  dsm_clear_status(voidhandle); /* clear status flags */

  /* Wait until we have both SIDEA and SIDEB data, completing a pair.  */
  totalpacks=0;
  while (!((handle->status & DSM_SIDEA) && (handle->status & DSM_SIDEB))) { 
    packs = dsm_update(handle, 2); /* check for new data */
    totalpacks += packs;
  }

  return totalpacks;
}


int dsm_get_unique_statuspair(DSMHandle voidhandle) {
  dsmhandle * handle = (dsmhandle *) voidhandle;
  int packs, totalpacks;

  if (handle == NULL)
    return DSM_ERR_BADHANDLE;

  dsm_clear_status(voidhandle); /* clear status flags */

  /* Wait until we have a SIDEA reading, ignoreing any SIDEB data.  */
  /* This will force us to sync on the next new pair of A/D values. */
  totalpacks=0;
  while (!(handle->status & DSM_SIDEA)) { 
    packs = dsm_update(handle, 1); /* check for new data */

    if (!(handle->status & DSM_SIDEA))
      dsm_clear_status(voidhandle);  /* clear status flags */

    rt_sleep(10); /* don't hog the CPU */
    totalpacks += packs;
  }

  /* Wait until we have both SIDEA and SIDEB data, completing this pair.  */
  while (!((handle->status & DSM_SIDEA) && (handle->status & DSM_SIDEB))) { 
    packs = dsm_update(handle, 1); /* check for new data */
    rt_sleep(10); /* don't hog the CPU */
    totalpacks += packs;
  }

  return totalpacks;
}


int dsm_get_unique_statusarray(DSMHandle voidhandle, 
                               int numsamples, int *range,
                               int *sidea, int *sideb) {
  dsmhandle * handle = (dsmhandle *) voidhandle;
  int i;
  for (i=0; i<numsamples; i++) {
    dsm_get_unique_statuspair(voidhandle);
    sidea[i] = handle->skyaduA;
    sideb[i] = handle->skyaduB;
  }
  *range = handle->ddcrangeA;

  return DSM_ERR_NONE;
}


int dsm_print_aperture_status(DSMHandle voidhandle) {
  dsmhandle * handle = (dsmhandle *) voidhandle;
  char wheelpos[8];

  if (!(handle->status & DSM_DATA))
    return DSM_ERR_NONE;

  switch (handle->aperture) {
    case DSM_APERTURE_IRFILTER: 
      strcpy(wheelpos, "IR ");
      break; 

    case DSM_APERTURE_BLACK: 
      strcpy(wheelpos, "Blk");
      break; 

    case DSM_APERTURE_OPEN: 
      strcpy(wheelpos, "Opn");
      break; 

    case DSM_APERTURE_WHITE: 
      strcpy(wheelpos, "Wht");
      break; 

    default:
      strcpy(wheelpos, "Unk");
      break; 
  }

  if (handle->apertureunknown) 
    strcpy(wheelpos, "Unk");

  if (handle->aperturemoving)
    strcpy(wheelpos, "Mov");

  printf("\rAperture wheel pos[%2d] filter:%s cnt:%3d", handle->aperture, wheelpos, handle->lastcount);
  fflush(stdout);

  return DSM_ERR_NONE;
}


int dsm_print_status(DSMHandle voidhandle) {
  dsmhandle * handle = (dsmhandle *) voidhandle;
  char wheelpos[8];

  if (!(handle->status & DSM_DATA))
    return DSM_ERR_NONE;

  switch (handle->aperture) {
    case DSM_APERTURE_IRFILTER: 
      strcpy(wheelpos, "IR ");
      break; 

    case DSM_APERTURE_BLACK: 
      strcpy(wheelpos, "Blk");
      break; 

    case DSM_APERTURE_OPEN: 
      strcpy(wheelpos, "Opn");
      break; 

    case DSM_APERTURE_WHITE: 
      strcpy(wheelpos, "Wht");
      break; 

    default:
      strcpy(wheelpos, "Unk");
      break; 
  }

  if (handle->apertureunknown) 
    strcpy(wheelpos, "Unk");

  if (handle->aperturemoving)
    strcpy(wheelpos, "Mov");

  printf("Status:[%2x] range:%d skyA:%7d skyB:%7d w[%2d]:%s temp: %.1f cnt:%3d\n",
         handle->status, handle->ddcrangeA,
         handle->skyaduA, handle->skyaduB, 
         handle->aperture, wheelpos, 
         handle->tempadu / 32.0, handle->lastcount);

  return DSM_ERR_NONE;
}



