/*
 * main.c - main program
 *
 * $Id: main.c,v 1.21 2008/08/30 10:06:31 johns Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <signal.h>

#include "util.h"
#include "serial.h"
#include "gps.h"
#include "dsm.h"
#include "sqmclient.h"
#include "info.h"
#include "calcstats.h"

/*
 * Handle SIGINT, SIGHUP, etc for graceful shutdown to safe operating mode
 */
/* global variable to indicate shutdown due to sigint on Unix */
int globalinterrupt;
int globalusr1trigger;
int globalusr2trigger;

void sigshutdownhandler(int sig) {
  globalinterrupt=1; 
}

void sigusr1handler(int sig) {
  globalusr1trigger=1; 
}

void sigusr2handler(int sig) {
  globalusr2trigger=1; 
}

void setsighandler(void) {
  globalinterrupt=0;
  globalusr1trigger=0;
  globalusr2trigger=0;

  signal(SIGINT, sigshutdownhandler);
#if !defined(_MSC_VER)
  /* Windows doesn't have SIGHUP, SIGUSR1, or SIGUSR2 */
  signal(SIGHUP,  sigshutdownhandler);
  signal(SIGUSR1, sigusr1handler);
  signal(SIGUSR2, sigusr2handler);
#endif
}


typedef struct {
  DSMHandle dsm;        /* DSM handle                                     */
  GPSHandle gps;        /* GPS handle                                     */
  char *dsmportname;    /* serial port device name for DSM                */
  char *gpsportname;    /* serial port device name for GPS                */
  char *sqmhostname;    /* internet address or hostname for SQM-LE        */
  int sqmport;          /* TCP port number for SQM-LE                     */
  double maxruntime;    /* maximum time to take readings before shutdown  */
  double sleepdelay;    /* sleep delay between I/O checks (sec)           */
  double timeout;       /* max elapsed time before DSM/GPS timeouts (sec) */ 
  int statsize;         /* number of samples to use for stats             */
  int aperture;         /* which aperture/filter to use for sampling      */
  char *tracklogname;   /* track log filename                             */
  FILE *tracklogfp;     /* track log file pointer                         */
} dsmgpsoptions;


/*
 * Print program usage info 
 */
int printusage(const char *exename) {
  printf("Usage:\n");
  printf(" DSM-only mode:  %s -dsmport dsmportname\n", exename);
  printf(" GPS-only mode:  %s -gpsport gpsportname\n", exename);
  printf(" SQM-only mode:  %s -sqmhost sqmhostname\n", exename);
  printf("  SQM+GPS mode:  %s -sqmhost sqmhostname -gpsport gpsportname\n", exename);
  printf("  DSM+GPS mode:  %s -dsmport dsmportname -gpsport gpsportname\n", exename);
  printf("     track log:  %s [...] -tracklog filename\n", exename);
  return 0;
}

/*
 * Initialize options to defaults
 */
void initoptions(dsmgpsoptions *opts) {
  memset(opts, 0, sizeof(dsmgpsoptions));
 
  /* DSM */
  /* Use IR blocking filter by default */
  opts->aperture = DSM_APERTURE_IRFILTER;
  opts->statsize = 10;

  /* SQM */
  opts->sqmport  = 10001; /* Standard TCP port used by device */
}


/* 
 * Parse next command line argument
 */
int getparm(int argc, char **argv, int num, dsmgpsoptions *opts) {
  if (!strcmp(argv[num], "-dsmport")) {
    opts->dsmportname = argv[num+1];
    return 2;
  }
  if (!strcmp(argv[num], "-gpsport")) {
    opts->gpsportname = argv[num+1];
    return 2;
  }
  if (!strcmp(argv[num], "-sqmhost")) {
    opts->sqmhostname = argv[num+1];
    return 2;
  }
  if (!strcmp(argv[num], "-sqmport")) {
    opts->sqmport     = atoi(argv[num+1]);
    return 2;
  }
  if (!strcmp(argv[num], "-tracklog")) {
    opts->tracklogname = argv[num+1];
    return 2;
  }

  return -1;
}


/*
 * Parse command line arguments and set state variables used in the
 * main program
 */
int parseargs(int argc, char **argv, dsmgpsoptions *opts) {
  int i, rc;

  if (opts == NULL)
    return -1;

  if (argc < 2) {
    printusage(argv[0]);
    return -1;
  }

  initoptions(opts); 

  i = 1;
  while (i < argc) {
    if (argv[i][0] == '-' || argv[i][0] == '+') {
      rc = getparm(argc, argv, i, opts);
      if (rc != -1) {
        i += rc;
      } else {
        printusage(argv[0]);
        return -1;
      }
    }
  }

  return 0;
}


/*
 * Take a sequence of unique readings storing them in the 
 * array provided by the caller
 */
int dsm_stats_run(dsmgpsoptions *opts) {
  if (opts->statsize > 0) {
    int *sidea, *sideb, range;
    double min, max, mean, rms, stddev;

    sidea = (int *) malloc(opts->statsize * sizeof(int));
    sideb = (int *) malloc(opts->statsize * sizeof(int));

    printf("Taking %d samples for stats processing...\n", opts->statsize);
    if (dsm_get_unique_statusarray(opts->dsm, opts->statsize, 
                               &range, sidea, sideb) == DSM_ERR_NONE) {
#if 0
      int i;
      for (i=0; i<opts->statsize; i++)
        printf("range: %d a: %d  b: %d\n", range, sidea[i], sideb[i]);
#endif
      calcstats_int(opts->statsize, sidea, &min, &max, &mean, &rms, &stddev);
      printf("SideA min: %.1f max: %.1f mean: %.1f stddev %f\n",
             min, max, mean, stddev);

      calcstats_int(opts->statsize, sideb, &min, &max, &mean, &rms, &stddev);
      printf("SideB min: %.1f max: %.1f mean: %.1f stddev %f\n",
             min, max, mean, stddev);
    }

    free(sidea);
    free(sideb);
  }

  return DSM_ERR_NONE;
}


/*
 * Main program
 */
int main(int argc, char **argv) {
  dsmgpsoptions opts;
  rt_timerhandle sqmtimer;

  printf("DSM/SQM/GPS sky brightness logging tool version 0.4\n");
  printf("by John E. Stone <john.stone@gmail.com>            \n");
  printf("---------------------------------------------------\n");

  if (parseargs(argc, argv, &opts))
    return -1;

  setsighandler(); /* capture interrupt signals and shutdown */
  sqmtimer = rt_timer_create();
  rt_timer_start(sqmtimer);

  /* 
   * Bring up GPS if available  
   *   The GPS should be brought up first, since it is a low baud rate
   *   device with unidirectional communication that we can safely 
   *   ignore for a while during the potentially lengthy DSM 
   *   initialization and configuration phase.
   */
  if (opts.gpsportname) {
    printf("Opening GPS port: %s\n", opts.gpsportname);

    if ((opts.gps = gps_open(opts.gpsportname)) == NULL) {
      printf("Failed to open GPS device\n");
      return -1;
    }

    if (!get_gps_fix(opts.gps, GPS_TIMELOC)) {
      int sysgpsdiff;

      print_gps_status(opts.gps);

      if (!gps_systime_diff(opts.gps, &sysgpsdiff)) {
        printf("System clock is %d seconds %s GPS time\n", sysgpsdiff,
               (sysgpsdiff > 0) ? "ahead of" : "behind");
      }
    } else {
      printf("Failed to get GPS UTC time.\n");
    }
  }


  /*
   * Open DSM port and attach to the device
   */
  if (opts.dsmportname) {
    printf("Opening DSM port: %s\n", opts.dsmportname);

    if ((opts.dsm = dsm_open(opts.dsmportname, 1)) == NULL) {
      printf("Failed to open DSM device\n");
      return -1;
    }
  } 


  /*
   * Initialize the DSM and set it to a known state
   *   We initialize the DSM last because it is a high baud rate device
   * that can't be squelched, throttled or otherwise ignored once we
   * start talking to it, otherwise we'll get dropped packets etc.
   */
  if (opts.dsm) {
    printf("Resetting DSM..\n");
    dsm_reset(opts.dsm);

    printf("Setting integration time...\n");
    dsm_set_integration_count(opts.dsm, 1000);
  }

  if (opts.tracklogname) {
    if (!opts.dsm && !opts.gps && !opts.sqmhostname) {
      printf("Cannot open a tracklog without one or more devices\n");
      return -1;
    } 
    opts.tracklogfp = fopen(opts.tracklogname, "a");
    if (opts.tracklogfp == NULL) {
      printf("Failed to open track log file `%s`, aborting\n", opts.tracklogname);
      return -1;
    }
  }

  /*
   * Begin exposure and data collection sequence
   */
  if (opts.dsm || (opts.sqmhostname != NULL) || (opts.gps != NULL)) {
    int totaldsmpackets=0;
    int sqmhavedata=0;    /* SQM data availability flag */
    sqm_reading sqmrd;    /* SQM reading */
    int gpshavedata=0;    /* GPS data availability flags */
    char datebuf[80];     /* GPS date buffer */
    char timebuf[80];     /* GPS time buffer */
    float declat, declon; /* GPS lat/long */
     
    if (opts.dsm) {
      printf("Commanding aperture wheel to requested filter position...\n");
      dsm_set_aperture(opts.dsm, opts.aperture);
    }

    printf("Starting exposure loop...\n");

    /* get first SQM reading with no timer check */
    sqmhavedata = (sqmclient_get_reading(opts.sqmhostname, opts.sqmport, &sqmrd) == 0);
    while (!globalinterrupt) {
      /*
       * DSM handling
       */
      if (opts.dsm) {
        int dsmpackets;
        if (opts.gps != NULL) {
          /*
           * Eat all of the pending buffered data so we get a reading that's
           * as close to the GPS data as possible
           */
          dsm_syncup_status(opts.dsm);
          rt_sleep(50); // don't chew CPU time unecessarily
  
          /*
           * Eat all of the pending buffered data so we get a reading that's
           * as close to the GPS data as possible
           */
          dsm_syncup_status(opts.dsm);
        } else {
          rt_sleep(50); // don't chew CPU time unecessarily
        }

        dsmpackets = dsm_get_unique_statuspair(opts.dsm);
        totaldsmpackets+=dsmpackets;
        if (dsmpackets > 0) {
          dsm_print_status(opts.dsm);
          dsm_clear_status(opts.dsm);
        }

        dsm_stats_run(&opts);
      }


      /* 
       * SQM-LE handling
       */
      if (opts.sqmhostname != NULL) {
        /*
         * Don't take readings faster than twice per second.
         * The max rate is really something close to 5 per second,
         * or we'll overload the device, but we want to be able to 
         * share the SQM with other apps/users.
         */
        if (rt_timer_timenow(sqmtimer) > 0.5) {
          sqmhavedata = (sqmclient_get_reading(opts.sqmhostname, opts.sqmport, &sqmrd) == 0);
          if (sqmhavedata) {
            printf("mag: %.2f ", sqmrd.mag_arcsec2); 
            printf("freq: %d ", sqmrd.frequency); 
            printf("period: %d  ", sqmrd.period); 
            printf("periodsec: %.3f  ", sqmrd.periodsec); 
            printf("temp: %.2f\n", sqmrd.temperature); 
          }
          rt_timer_start(sqmtimer);
        }
      }


      /*
       * User-defined signal handler usage
       */
      if (globalusr1trigger) {
        printf("Received signal, initiating sequence of 10 readings..\n");
        if (opts.dsm)
          dsm_stats_run(&opts);
      } 
      if (globalusr2trigger) {
        printf("Received signal, initiating usr2 operation..\n");
      }


      /*
       * GPS handling
       */
      if (opts.gps != NULL) {
        get_gps_fix(opts.gps, GPS_TIMELOC);
        gps_have_data(opts.gps, &gpshavedata);
        if (gpshavedata & GPS_COMM) {
          printf("GPS ");
          if (gpshavedata & (GPS_DATE | GPS_TIME)) {
            printf("UTC ");
          }
          if (gpshavedata & GPS_DATE) {
            gps_sprint_date_utc(opts.gps, datebuf);
            printf("date: %s ", datebuf);
          }
          if (gpshavedata & GPS_TIME) {
            gps_sprint_time_utc(opts.gps, timebuf);
            printf("time: %s ", timebuf);
          }
          if (gpshavedata & (GPS_FIXLAT | GPS_FIXLONG)) {
            gps_get_lat_decimal(opts.gps, &declat);
            gps_get_long_decimal(opts.gps, &declon);
            printf("lat: %f long: %f ", declat, declon);
          }

          printf("\n");
        }
      }

      if (opts.tracklogfp) {
        if (gpshavedata & (GPS_COMM | GPS_DATE | GPS_TIME | GPS_FIXLAT | GPS_FIXLONG)) {
          fprintf(opts.tracklogfp, "%s,%s,%.4f,%.4f", datebuf, timebuf, declat, declon);
        } else {
          fprintf(opts.tracklogfp, ",,,");
        }

        if (sqmhavedata) {
          fprintf(opts.tracklogfp, ",%.2f", sqmrd.mag_arcsec2);
        } else {
          fprintf(opts.tracklogfp, ",");
        }
        fprintf(opts.tracklogfp, "\n");
        fflush(opts.tracklogfp);
      }

    } /* End of main loop */

    /*
     * Graceful shutdown, restoring the equipment to safe mode of operation
     */
    printf("Received signal, initiating gracefull shutdown.\n");

    if (opts.dsm) {
      printf("DSM safe shutdown in progress.\n");
      printf("Commanding aperture wheel to the white heat blocking position.\n");
      dsm_set_aperture(opts.dsm, DSM_APERTURE_WHITE);
    } 
  } 

  if (opts.dsm)
    dsm_close(opts.dsm);

  if (opts.gps)
    gps_close(opts.gps);

  if (opts.tracklogfp != NULL) {
    printf("Closing track log file `%s`\n", opts.tracklogname);
    fclose(opts.tracklogfp);
  }

  rt_timer_destroy(sqmtimer);
 
  return 0; 
}


