#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "sock.h"
#include "sqmclient.h"

void *sqmclient_open_socket(const char *host, int port) {
  void *sock;
  
  sock_init(); /* make sure Winsock is initialized */
  sock = sock_create();
  if (sock == NULL) {
    printf("Failed to connect, couldn't create socket\n");
    return NULL;
  }
 
  if (sock_connect(sock, host, port) < 0) {
    printf("Failed to connect to SQM-LE host `%s', port %d\n", host, port);
    sock_destroy(sock);
    return NULL;
  }

  return sock;
}


static int sock_readn(void *s, char *ptr, int n) {
  int nleft;
  int nread;

  nleft = n;
  while (nleft > 0) {
    if ((nread = sock_read(s, ptr, nleft)) < 0) {
      if (errno == EINTR)
        nread = 0;         /* and call read() again */
      else
        return -1;
    } else if (nread == 0)
      break;               /* EOF */
    nleft -= nread;
    ptr += nread;
  }
  return n-nleft;
}


static int sock_writestr(void *sock, const char *str) {
  int len = strlen(str);
  int nleft;
  int nwritten;

  nleft = len;
  while (nleft > 0) {
    if ((nwritten = sock_write(sock, str, nleft)) <= 0) {
      if (errno == EINTR)
        nwritten = 0;
      else
        return -1;
    }
    nleft -= nwritten;
    str += nwritten;
  }
  return len;
}


int sqmclient_get_reading(const char *host, int port, sqm_reading *reading) {
  void *sock = sqmclient_open_socket(host, port);
  char readingstr[80];
  int scancount;

  if (!sock) {
    return -1;
  }  

  /* wait up to 5 seconds for connection activity */ 
  if (sock_selwrite(sock, 5) != 1) return -1;
 
  sock_writestr(sock, "rx\n");
  memset(readingstr, 0, sizeof(readingstr));
  sock_readn(sock, readingstr, 57);
  sock_destroy(sock);

  readingstr[55]='\0';
  readingstr[56]='\0';

#if 0
  printf("Reading: '%s'\n", readingstr);
#endif

  scancount = sscanf(readingstr, "r,%fm,%dHz,%dc,%fs,%fC\r\n",
                     &reading->mag_arcsec2, 
                     &reading->frequency, 
                     &reading->period,
                     &reading->periodsec,
                     &reading->temperature);

  if (scancount != 5) {
    printf("Failed to parse SQM output\n");
    return -1;
  }

  return 0;
}

int sqmclient_print_reading(const char *host, int port) {
  sqm_reading r;

  if (sqmclient_get_reading(host, port, &r) != 0)
    return -1;

#if 1
  printf("mag: %.2f ", r.mag_arcsec2);
  printf("freq: %d ", r.frequency);
  printf("period: %d  ", r.period);
  printf("periodsec: %.3f  ", r.periodsec);
  printf("temp: %.2f\n", r.temperature);
#else
  printf("reading:\n");
  printf("  mag/arcsec^2: %.2f\n", r.mag_arcsec2);
  printf("     frequency: %d\n", r.frequency);
  printf("        period: %d\n", r.period);
  printf("     periodsec: %.3f\n", r.periodsec);
  printf("   temperature: %.2f\n", r.temperature);
#endif
  return 0;
}


